from dataclasses import dataclass, field
from datetime import datetime
from typing import List, Optional, Union

from dataclasses_json.api import dataclass_json
from dataclasses_json.cfg import config
from stringcase import camelcase


def datetime_encoder(dt: Optional[datetime] = None) -> Optional[str]:
    if dt is None:
        return None

    return dt.isoformat()


@dataclass_json(letter_case=camelcase)
@dataclass
class Browser:
    name: str
    version: str
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Creator:
    name: str
    version: str
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class CacheState:
    last_access: str
    e_tag: str
    hit_count: int
    expires: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Cache:
    before_request: Optional[CacheState] = None
    after_request: Optional[CacheState] = None
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class QueryParameter:
    name: str
    value: str


@dataclass_json(letter_case=camelcase)
@dataclass
class Header:
    name: str
    value: str


@dataclass_json(letter_case=camelcase)
@dataclass
class Cookie:
    name: str
    value: str
    path: Optional[str] = None
    domain: Optional[str] = None
    expires: Optional[datetime] = field(
        default=None,
        metadata=config(encoder=datetime_encoder),
    )
    http_only: Optional[bool] = None
    secure: Optional[bool] = None
    same_site: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Param:
    name: str
    value: Optional[str] = None
    file_name: Optional[str] = None
    content_type: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class PostData:
    mime_type: str
    params: List[Param]
    text: str


@dataclass_json(letter_case=camelcase)
@dataclass
class Request:
    method: str
    url: str
    http_version: str
    cookies: List[Cookie]
    headers: List[Header]
    query_string: List[QueryParameter]
    headers_size: int
    body_size: int
    post_data: Optional[PostData] = None
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Content:
    size: int
    compression: Optional[int] = None
    mime_type: Optional[str] = None
    text: Optional[str] = None
    encoding: Optional[str] = None
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Response:
    status: int
    status_text: str
    http_version: str
    cookies: List[Cookie]
    headers: List[Header]
    content: Content
    headers_size: int
    body_size: int
    redirect_url: str = field(metadata=config(field_name="redirectURL"))
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Timings:
    send: Union[int, float]
    wait: Union[int, float]
    receive: Union[int, float]
    blocked: Optional[Union[int, float]] = None
    dns: Optional[Union[int, float]] = None
    connect: Optional[Union[int, float]] = None
    ssl: Optional[Union[int, float]] = None
    comment: Optional[Union[int, float]] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Entry:
    started_date_time: datetime = field(
        metadata=config(
            encoder=datetime_encoder,
        )
    )
    time: Union[int, float]
    request: Request
    response: Response
    cache: Cache
    timings: Timings
    pageref: Optional[str] = None
    server_ip_address: Optional[str] = field(
        default=None, metadata=config(field_name="serverIPAddress")
    )
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class PageTimings:
    on_content_load: Union[int, float, None] = None
    on_load: Union[int, float, None] = None
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Page:
    started_date_time: datetime = field(metadata=config(encoder=datetime_encoder))
    id: str
    title: str
    page_timings: PageTimings
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Log:
    version: str
    creator: Creator
    browser: Browser
    pages: List[Page]
    entries: List[Entry]
    comment: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class Har:
    log: Log



from dataclasses import dataclass, field
from typing import Any, Dict, Optional

from dataclasses_json.api import dataclass_json
from dataclasses_json.cfg import config
from stringcase import camelcase


@dataclass_json(letter_case=camelcase)
@dataclass
class Response:
    url: str
    status: int
    status_text: str
    headers: Dict[str, Any]
    mime_type: str
    connection_reused: bool
    connection_id: int
    encoded_data_length: int
    security_state: str
    response_time: Optional[float] = None
    request_headers: Optional[Dict[str, Any]] = None
    request_headers_text: Optional[str] = None
    remote_ip_address: Optional[str] = field(
        default=None, metadata=config(field_name="remoteIPAddress")
    )
    remote_port: Optional[int] = None
    from_disk_cache: Optional[bool] = None
    from_service_worker: Optional[bool] = None
    from_prefetch_cache: Optional[bool] = None
    timing: Optional[Dict[str, float]] = None
    protocol: Optional[str] = None
    headers_text: Optional[str] = None


@dataclass_json(letter_case=camelcase)
@dataclass
class ResponseReceivedEvent:
    request_id: str
    loader_id: str
    timestamp: float
    type: str
    response: Response
    frame_id: str



import re
from datetime import datetime, timedelta
from typing import Dict, List, Optional
from urllib.parse import parse_qs

import dateutil.parser
from playwright.async_api import Request

from . import dataclasses


def millis_to_roundish_millis(value: float) -> int:
    return int(int(value * 1000) / 1000)


def datetime_to_millis(dt: datetime) -> float:
    return dt.timestamp() * 1000.0


def query_to_query_params(query: str) -> List[dataclasses.har.QueryParameter]:
    query_params: List[dataclasses.har.QueryParameter] = []

    parsed_query_params = parse_qs(query)
    for name, values in parsed_query_params.items():
        value = "".join(values)
        query_params.append(dataclasses.har.QueryParameter(name=name, value=value))

    return query_params


def form_data_to_params(data: str) -> List[dataclasses.har.Param]:
    params = query_to_query_params(data)
    return [
        dataclasses.har.Param(name=param.name, value=param.value) for param in params
    ]


def dict_to_headers(dict_: Dict[str, str]) -> List[dataclasses.har.Header]:
    return [
        dataclasses.har.Header(name=name, value=value) for name, value in dict_.items()
    ]


def parse_cookie(c: str) -> dataclasses.har.Cookie:
    cookie = dataclasses.har.Cookie(name="", value="")

    first = True
    for pair in re.split(r"; *", c):
        pair = str(pair)

        index_of_equals = -1
        if "=" in pair:
            index_of_equals = pair.index("=")

        name = pair[0:index_of_equals] if index_of_equals != -1 else pair.strip()
        value = pair[index_of_equals + 1 :] if index_of_equals != -1 else ""

        if first:
            first = False
            cookie.name = name
            cookie.value = value
            continue

        if name == "Domain":
            cookie.domain = value

        if name == "Expires":
            cookie.expires = dateutil.parser.parse(value)

        if name == "HttpOnly":
            cookie.http_only = True

        if name == "Max-Age":
            max_age = datetime.now() + timedelta(seconds=int(value))
            cookie.expires = max_age

        if name == "Path":
            cookie.path = value

        if name == "SameSite":
            cookie.same_site = value

        if name == "Secure":
            cookie.secure = True

    return cookie


def cookies_for_har(
    header: Optional[str],
    separator: str,
) -> List[dataclasses.har.Cookie]:
    if header is None:
        return []

    return [parse_cookie(c) for c in header.split(separator)]


def post_data_for_har(request: Request) -> Optional[dataclasses.har.PostData]:
    post_data = request.post_data_buffer
    if post_data is None:
        return None

    content_type = request.headers.get("content-type", "application/octet-stream")
    text = "" if content_type == "application/octet-stream" else post_data.decode()
    result = dataclasses.har.PostData(mime_type=content_type, text=text, params=[])

    if content_type == "application/x-www-form-urlencoded":
        form_data = post_data.decode()
        result.params.extend(form_data_to_params(form_data))

    return result


def set_remote_ip_address_as_comment(
    har: dataclasses.har.Har,
    response_received_events: List[dataclasses.cdp.ResponseReceivedEvent],
):
    # url -> remote_ip_address table
    memo: Dict[str, Optional[str]] = {}
    for event in response_received_events:
        key = event.response.url
        value = event.response.remote_ip_address
        memo[key] = value

    # set a remote IP address as a comment
    for entry in har.log.entries:
        url = entry.request.url
        entry.response.comment = memo.get(url)

    return har



import asyncio
import base64
import copy
from datetime import datetime
from typing import Dict, List, Union
from urllib.parse import urlparse

import poetry_version
from playwright.async_api import BrowserContext, CDPSession, Page, Request, Response

from . import dataclasses
from .utils import (
    cookies_for_har,
    datetime_to_millis,
    dict_to_headers,
    millis_to_roundish_millis,
    post_data_for_har,
    query_to_query_params,
    set_remote_ip_address_as_comment,
)

__version__ = poetry_version.extract(source_file=__file__)


HAR_VERSION: str = "1.2"
CREATOR_NAME: str = "playwright-har-tracer"
CREATOR_VERSION: str = __version__


class HarTracer:
    def __init__(
        self,
        context: BrowserContext,
        browser_name: str,
        *,
        omit_content: bool = False,
    ):
        if context.browser is None:
            raise ValueError

        self._omit_content = omit_content

        self._page_entries: Dict[Page, dataclasses.har.Page] = {}
        self._entries: Dict[Request, dataclasses.har.Entry] = {}
        self._last_page: int = 0

        self._loop = asyncio.get_event_loop()
        self._tasks: List[asyncio.Task] = []

        self._on_load_event = asyncio.Event()
        self._on_dom_content_loaded_event = asyncio.Event()

        self._log = dataclasses.har.Log(
            version=HAR_VERSION,
            creator=dataclasses.har.Creator(name=CREATOR_NAME, version=CREATOR_VERSION),
            browser=dataclasses.har.Browser(
                name=browser_name, version=context.browser.version
            ),
            pages=[],
            entries=[],
        )

        self._response_received_events: List[dataclasses.cdp.ResponseReceivedEvent] = []

        def on_request(page: Page, request: Request) -> None:
            page_entry = self._page_entries.get(page)
            if page_entry is None:
                return

            parsed_url = urlparse(request.url)

            har_entry = dataclasses.har.Entry(
                pageref=page_entry.id,
                started_date_time=datetime.now(),
                time=-1,
                request=dataclasses.har.Request(
                    method=request.method,
                    url=request.url,
                    http_version="HTTP/1.1",
                    cookies=[],
                    headers=[],
                    query_string=query_to_query_params(parsed_url.query),
                    post_data=None,
                    headers_size=-1,
                    body_size=-1,
                ),
                response=dataclasses.har.Response(
                    status=-1,
                    status_text="",
                    http_version="HTTP/1.1",
                    cookies=[],
                    headers=[],
                    content=dataclasses.har.Content(
                        size=-1,
                        mime_type=request.headers.get("content-type")
                        or "application/octet-stream",
                    ),
                    headers_size=-1,
                    body_size=-1,
                    redirect_url="",
                ),
                cache=dataclasses.har.Cache(before_request=None, after_request=None),
                timings=dataclasses.har.Timings(send=-1, wait=-1, receive=-1),
            )

            redirected_from_request = request.redirected_from
            if redirected_from_request is not None:
                from_entry = self._entries.get(redirected_from_request)
                if from_entry is not None:
                    from_entry.response.redirect_url = request.url

            self._log.entries.append(har_entry)
            self._entries[request] = har_entry

        def on_response(page: Page, response: Response) -> None:
            page_entry = self._page_entries.get(page)
            if page_entry is None:
                return

            request = response.request
            har_entry = self._entries.get(request)
            if har_entry is None:
                return

            # Rewrite provisional headers with actual
            har_entry.request.headers = dict_to_headers(request.headers)
            har_entry.request.cookies = cookies_for_har(
                request.headers.get("cookie"), ";"
            )
            har_entry.request.post_data = post_data_for_har(request)

            har_entry.response = dataclasses.har.Response(
                status=response.status,
                status_text=response.status_text,
                http_version="HTTP/1.1",
                cookies=cookies_for_har(response.headers.get("set-cookie"), "\n"),
                headers=dict_to_headers(response.headers),
                content=dataclasses.har.Content(
                    size=-1,
                    mime_type=response.headers.get(
                        "content-type", "application/octet-stream"
                    ),
                ),
                headers_size=-1,
                body_size=-1,
                redirect_url="",
            )

            timing = response.request.timing
            start_time = timing.get("startTime", 0.0)
            if datetime_to_millis(page_entry.started_date_time) > start_time:
                page_entry.started_date_time = datetime.fromtimestamp(
                    start_time / 1000.0
                )

            domain_lookup_start: Union[float, int] = timing.get("domainLookupStart", -1)
            domain_lookup_end: Union[float, int] = timing.get("domainLookupEnd", -1)
            dns = (
                millis_to_roundish_millis(domain_lookup_end - domain_lookup_start)
                if domain_lookup_end != -1
                else -1
            )

            connect_start: Union[float, int] = timing.get("connectStart", -1)
            connect_end: Union[float, int] = timing.get("connectEnd", -1)
            connect = (
                millis_to_roundish_millis(connect_end - connect_start)
                if connect_end != -1
                else -1
            )

            secure_connection_start: Union[float, int] = timing.get(
                "secureConnectionStart", -1
            )
            ssl = (
                millis_to_roundish_millis(connect_end - secure_connection_start)
                if connect_end != -1
                else -1
            )

            request_start: Union[float, int] = timing.get("requestStart", -1)
            response_start: Union[float, int] = timing.get("responseStart", -1)
            wait = (
                millis_to_roundish_millis(response_start - request_start)
                if response_start != -1
                else -1
            )

            response_end: Union[float, int] = timing.get("responseEnd", -1)
            receive = (
                millis_to_roundish_millis(response_end - response_start)
                if response_end != -1
                else -1
            )

            har_entry.timings = dataclasses.har.Timings(
                dns=dns,
                connect=connect,
                ssl=ssl,
                send=0,
                wait=wait,
                receive=receive,
            )
            har_entry.time = sum([dns, connect, ssl, wait, receive])

            if self._omit_content is False and response.status == 200:

                async def on_response_task():
                    body = await response.body()
                    har_entry.response.content.text = base64.b64encode(body).decode()
                    har_entry.response.content.encoding = "base64"

                self._tasks.append(self._loop.create_task(on_response_task()))

        def on_page(page: Page) -> None:
            page_entry = dataclasses.har.Page(
                started_date_time=datetime.now(),
                id=f"page_{self._last_page}",
                title="",
                page_timings=dataclasses.har.PageTimings(
                    on_content_load=-1, on_load=-1
                ),
            )
            self._last_page += 1

            self._page_entries[page] = page_entry
            self._log.pages.append(page_entry)

            page.on("request", lambda request: on_request(page, request))
            page.on("response", lambda response: on_response(page, response))

            def on_dom_content_loaded(page: Page) -> None:
                async def on_dom_content_loaded_task():
                    result: Dict[str, Union[str, int]] = await page.main_frame.evaluate(
                        """Promise.resolve({title: document.title, domContentLoaded: performance.timing.domContentLoadedEventStart})"""
                    )

                    title = str(result.get("title", ""))
                    page_entry.title = title

                    dom_content_loaded = int(result.get("domContentLoaded", 0))
                    page_entry.page_timings.on_content_load = dom_content_loaded

                    self._on_dom_content_loaded_event.set()

                self._tasks.append(self._loop.create_task(on_dom_content_loaded_task()))

            async def wait_on_dom_content_loaded_task():
                await self._on_dom_content_loaded_event.wait()

            self._tasks.append(
                self._loop.create_task(wait_on_dom_content_loaded_task())
            )

            def on_load(page: Page) -> None:
                async def on_load_task():
                    result: Dict[str, Union[str, int]] = await page.main_frame.evaluate(
                        """Promise.resolve({title: document.title, loaded: performance.timing.loadEventStart})"""
                    )

                    title = str(result.get("title", ""))
                    page_entry.title = title

                    loaded = int(result.get("loaded", 0))
                    page_entry.page_timings.on_load = loaded

                    self._on_load_event.set()

                self._tasks.append(self._loop.create_task(on_load_task()))

            async def wait_on_load_task():
                await self._on_load_event.wait()

            self._tasks.append(self._loop.create_task(wait_on_load_task()))

            page.on("domcontentloaded", lambda: on_dom_content_loaded(page))
            page.on("load", lambda: on_load(page))

        context.on("page", on_page)

    async def enable_response_received_event_tracing(self, client: CDPSession):
        await client.send("Network.enable")
        client.on(
            "Network.responseReceived",
            lambda data: self._response_received_events.append(
                dataclasses.cdp.ResponseReceivedEvent.from_dict(data)
            ),
        )

    async def flush(self) -> dataclasses.har.Har:
        await asyncio.gather(*self._tasks)

        for page_entry in self._log.pages:
            on_content_load = page_entry.page_timings.on_content_load
            if on_content_load is not None and float(on_content_load) >= 0.0:
                page_entry.page_timings.on_content_load = float(
                    on_content_load
                ) - datetime_to_millis(page_entry.started_date_time)
            else:
                page_entry.page_timings.on_content_load = -1

            on_load = page_entry.page_timings.on_load
            if on_load is not None and float(on_load) >= 0.0:
                page_entry.page_timings.on_load = float(on_load) - datetime_to_millis(
                    page_entry.started_date_time
                )
            else:
                page_entry.page_timings.on_load = -1

        log = copy.deepcopy(self._log)
        har = dataclasses.har.Har(log=log)
        har = set_remote_ip_address_as_comment(har, self._response_received_events)
        return har



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['playwright_har_tracer', 'playwright_har_tracer.dataclasses']

package_data = \
{'': ['*']}

install_requires = \
['dataclasses-json>=0.5.3,<0.6.0',
 'poetry-version>=0.1.5,<0.2.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'stringcase>=1.2.0,<2.0.0']

setup_kwargs = {
    'name': 'playwright-har-tracer',
    'version': '0.1.0',
    'description': "A Python implementation version of Playwright's HAR tracer",
    'long_description': '# playwright-har-tracer\n\nA Python implementation version of Playwright\'s HAR tracer.\n\n## Motivation\n\nPlaywright\'s HAR tracer is implemented to generate HAR as a file. I need to get HAR as a Python object rather than a file.\n\n- `playwright-har-tracer`\'s HarTracer generates HAR as a dataclass object.\n\n## ⚠️ Limitations\n\n- Tested with Python 3.8+\n- Tested with Chromium only\n- Supports the async API only\n\n## Installation\n\n```bash\npip install playwright-har-tracer\n```\n\n## Usage\n\n```python\nimport asyncio\nfrom playwright.async_api import async_playwright\nfrom playwright_har_tracer import HarTracer\n\n\nasync def main():\n    async with async_playwright() as p:\n        browser = await p.chromium.launch()\n        context = await browser.new_context()\n\n        tracer = HarTracer(context=context, browser_name=p.chromium.name)\n\n        page = await context.new_page()\n\n        await page.goto("http://whatsmyuseragent.org/")\n\n        har = await tracer.flush()\n\n        await context.close()\n        await browser.close()\n\n    print(har.to_json())\n\n\nasyncio.run(main())\n```\n',
    'author': 'Manabu Niseki',
    'author_email': 'manabu.niseki@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ninoseki/playwright-har-tracer',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
